#include<ctime>
#include<cmath>
#include<cstdio>
#include<cstring>
#include<cstdlib>
#include<cassert>
#include<string>
#include<iostream>
#include<fstream>
#include<vector>
#include<algorithm>

#include "PosPair.h"
#include "bcTree.h"

using namespace std;

const int STRLEN = 1005;

struct Alignment {
  char cid, dir;
  int pos, rid;
  double frac;

  Alignment() { cid = dir = 0; pos = rid = 0; frac = 0.0; }

  Alignment(char cid, char dir, int pos, int rid, double frac = 0.0) {
    this->cid = cid;
    this->dir = dir;
    this->pos = pos;
    this->rid = rid;
    this->frac = frac;
  }
};

int ROUND = 0;
int UPPERBOUND = 200; // 200 by default

int windowSize, halfws;
int m, n, ns; // m chromosomes, n reads, ns tot # of alignments
int *seqLen;

int nUniq, nMulti; //nUniq unique reads, nMulti multi reads
int nAligns; //# of alignments for multireads
vector<int> s;
vector<Alignment> uniq, multi;

bcTree **bctrees;

void loadRef(char * refF) {
  FILE *fi = NULL;

  //load ref file
  fi = fopen(refF, "r");
  assert((fscanf(fi, "%d", &m) == 1) && m > 0);
  seqLen = new int[m];
  for (int i = 0; i < m; i++) assert(fscanf(fi, "%d", &seqLen[i]) == 1);
  fclose(fi);
}

void loadData(istream &fin) {
  string line;
  int crid;
  char dir, cid;
  int pos, rid;
  char *num, str[STRLEN];
  vector<Alignment> aligns;


  n = nUniq = nMulti = 0;
  ns = nAligns = 0;
  s.clear(); uniq.clear(); multi.clear();

  crid = -1;
  aligns.clear();

  //load read file
  while (getline(fin, line)) {
    //parse results
    size_t pt = line.find(':');
    assert(pt != string::npos);
    dir = line[pt - 1];
    rid = atoi(line.substr(0, pt - 1).c_str());
    strcpy(str, line.substr(pt + 2).c_str());
    num = strtok(str, ",");
    assert(num != NULL);
    cid = atoi(num);    // chromosome id starts from 0
    assert((num = strtok(NULL, ",")) != NULL);
    pos = atoi(num);

    if (crid != rid) {
      if (crid >= 0) {
	if (aligns.size() == 1) {	
	  uniq.insert(uniq.end(), aligns.begin(), aligns.end());
	  ++nUniq;
	}
	else {
	  if (nAligns == 0) s.push_back(0);
	  multi.insert(multi.end(), aligns.begin(), aligns.end());
	  nAligns += aligns.size();
	  s.push_back(nAligns);
	  ++nMulti;
	}
      }
      crid = rid;
      aligns.clear();
    }

    aligns.push_back(Alignment(cid, dir, pos, rid));

    ++ns;
    if (ns % 1000000 == 0) printf("FIN %d\n", ns);
  }

  if (crid >= 0) {
    if (aligns.size() == 1) {	
      uniq.insert(uniq.end(), aligns.begin(), aligns.end());
      ++nUniq;
    }
    else {
      if (nAligns == 0) s.push_back(0);
      multi.insert(multi.end(), aligns.begin(), aligns.end());
      nAligns += aligns.size();
      s.push_back(nAligns);
      ++nMulti;
    }
  }

  n = nUniq + nMulti;
  printf("Load File Finished!\n");
  printf("nUniq = %d, nMulti = %d, n = %d\n", nUniq, nMulti, n);
}

void constructBCTrees() {
  int nPos;
  PosPair *arr = new PosPair[ns];

  for (int i = 0; i < nUniq; i++) {
    arr[i].cid = uniq[i].cid; arr[i].pos = uniq[i].pos;
  }
  for (int i = 0; i < nAligns; i++) {
    arr[i + nUniq].cid = multi[i].cid; arr[i + nUniq].pos = multi[i].pos;
  }
  sort(arr, arr + ns);
  nPos = unique(arr, arr + ns) - arr;

  if (m <= 0) { fprintf(stderr, "Number of chromosomes is less than 1!\n"); exit(-1); }

  bctrees = new bcTree*[m]; // you should guarantee that m > 0
  memset(bctrees, 0, sizeof(bcTree*) * m);

  int curpos = 0, curcid, curend;

  do {
    curcid = arr[curpos].cid;
    curend = curpos + 1;
    while (curend < nPos && curcid == arr[curend].cid) ++curend;
    bctrees[curcid] = new bcTree(curend - curpos, arr + curpos);
    curpos = curend;
  } while(curpos < nPos);

  for (int i = 0; i < m; i++) 
    if (bctrees[i] == NULL) {
      bctrees[i] = new bcTree();
    }

  delete[] arr;
  printf("Finish Construct BCTrees\n");
}

void EM() {
  double ll, ll_old;
  double genomeL;
  vector<double> rec;

  char cid;
  double tot;
  int seql, pos, lb, ub;

  genomeL = 0.0;
  for (int i = 0; i < m; i++) genomeL += seqLen[i];

  //Uniform Prior, initialize
  ll = -n * log(genomeL);
  for (int i = 0; i < nUniq; i++) {
    cid = uniq[i].cid; pos = uniq[i].pos; seql = seqLen[cid];
    lb = max(-1, pos - halfws - 1); ub = min(seql - 1, pos + halfws);
    ll += log(double(ub - lb));
    uniq[i].frac = 1.0;
  }
  for (int i = 0; i < nMulti; i++) {
    tot = 0.0; rec.clear();
    for (int j = s[i]; j < s[i + 1]; j++) {
      cid = multi[j].cid; pos = multi[j].pos; seql = seqLen[cid];
      lb = max(-1, pos - halfws - 1); ub = min(seql - 1, pos + halfws);
      rec.push_back(double(ub - lb));
      tot += (ub - lb);
    }
    ll += log(tot);
    for (int j = s[i]; j < s[i + 1]; j++) {
      multi[j].frac = rec[j - s[i]] / tot;
    }
  }

  printf("ll_initial = %.15g\n", ll);

  ROUND = 0;

  do {
    ll_old = ll;

    time_t a, b;
    long tcall;

    a = time(NULL);
    tinsert = tcall = 0; 

    //clear bctrees
    for (int i = 0; i < m; i++) bctrees[i]->clear();

    //M step  ??? Somehow M step @_@|||
    for (int i = 0; i < nUniq; i++) {
      tcall++;
      bctrees[uniq[i].cid]->insert(uniq[i].pos, uniq[i].frac);
    }
    for (int i = 0; i < nMulti; i++)
      for (int j = s[i]; j < s[i + 1]; j++) {
	tcall++;
	bctrees[multi[j].cid]->insert(multi[j].pos, multi[j].frac);
      }
    b = time(NULL);
    printf("%d %.4lf\n", int(b - a), tinsert / 1.0 / tcall);

    a = time(NULL);
    tcount = tcall = 0;

    //E step
    ll = 0.0;
    for (int i = 0; i < nUniq; i++) {
      cid = uniq[i].cid; pos = uniq[i].pos; seql = seqLen[cid];
      lb = max(-1, pos - halfws - 1); ub = min(seql - 1, pos + halfws);
      ll += log(bctrees[cid]->count(ub) - bctrees[cid]->count(lb));
      uniq[i].frac = 1.0;

      tcall += 2;
    }
    for (int i = 0; i < nMulti; i++) {
      tot = 0.0; rec.clear();
      for (int j = s[i]; j < s[i + 1]; j++) {
	cid = multi[j].cid; pos = multi[j].pos; seql = seqLen[cid];
	lb = max(-1, pos - halfws - 1); ub = min(seql - 1, pos + halfws);
	double val = bctrees[cid]->count(ub) - bctrees[cid]->count(lb);
	rec.push_back(val);
	tot += val;

	tcall += 2;
      }
      ll += log(tot);
      for (int j = s[i]; j < s[i + 1]; j++) {
	multi[j].frac = rec[j - s[i]] / tot;
      }
    }
    ll -= n * log(1.0 * n);

    b = time(NULL);
    printf("%d %.4lf\n", int(b - a), tcount / 1.0 / tcall);

    ++ROUND;
    printf("ROUND = %d, ll = %.15g, ll_old = %.15g\n", ROUND, ll, ll_old);
 
  } while(ll > ll_old + 0.1 && ROUND < UPPERBOUND);
}

void output(char* outF) {
  FILE *fo = fopen(outF, "w");
  fprintf(fo, "%d %d\n", nUniq, nMulti);
  for (int i = 0; i < nUniq; i++) {
    fprintf(fo, "%d\t%d %c %d %.15g\n", uniq[i].rid, uniq[i].cid, uniq[i].dir, uniq[i].pos, uniq[i].frac);
  }
  for (int i = 0; i < nAligns; i++) {
    fprintf(fo, "%d\t%d %c %d %.15g\n", multi[i].rid, multi[i].cid, multi[i].dir, multi[i].pos, multi[i].frac);
  }
  fclose(fo);
}

void freeAll() {
  delete[] seqLen;
  for (int i = 0; i < m; i++) {
    delete bctrees[i];
  }
  delete[] bctrees;
}

bool readFromFile;
char dataF[STRLEN];

int main(int argc, char* argv[]) {
  if (argc < 5) {
    printf("Usage : %s refF windowSize UPPERBOUND outF [-d dataF]\n", argv[0]);
    exit(0);
  }

  readFromFile = false;
  for (int i = 5; i < argc; i++) {
    if (!strcmp(argv[i], "-d")) {
      strcpy(dataF, argv[i + 1]);
      break;
    }
  }

  windowSize = atoi(argv[2]);
  assert(windowSize & 1);
  halfws = windowSize / 2;

  UPPERBOUND = atoi(argv[3]);

  loadRef(argv[1]);

  if (readFromFile) {
    ifstream fin;
    fin.open(dataF);
    loadData(fin);
    fin.close();
  }
  else {
    loadData(cin);
  }

  constructBCTrees(); // BCTrees, binary counting trees
  EM();
  output(argv[4]);
  freeAll();

  return 0;
}
